from great_expectations.datasource.fluent.data_asset.path.file_asset import FileDataAsset

class CSVAsset(FileDataAsset): ...
class ExcelAsset(FileDataAsset): ...
class JSONAsset(FileDataAsset): ...
class ORCAsset(FileDataAsset): ...
class ParquetAsset(FileDataAsset): ...
class FeatherAsset(FileDataAsset): ...
class FWFAsset(FileDataAsset): ...
class HDFAsset(FileDataAsset): ...
class HTMLAsset(FileDataAsset): ...
class PickleAsset(FileDataAsset): ...
class SASAsset(FileDataAsset): ...
class SPSSAsset(FileDataAsset): ...
class StataAsset(FileDataAsset): ...
class XMLAsset(FileDataAsset): ...
