from typing import List

from tests.integration.integration_test_fixture import IntegrationTestFixture

trino_integration_tests = []

connecting_to_your_data: List[IntegrationTestFixture] = []


trino_integration_tests += connecting_to_your_data
