COLUMN_DESCRIPTION = "The column name."
COLUMN_A_DESCRIPTION = "The first column name."
COLUMN_B_DESCRIPTION = "The second column name."
COLUMN_LIST_DESCRIPTION = "Set of columns to be checked."
MOSTLY_DESCRIPTION = "Successful if at least `mostly` fraction of values match the Expectation."
IGNORE_ROW_IF_DESCRIPTION = (
    "If specified, sets the condition on which a given row is to be ignored."
)
VALUE_SET_DESCRIPTION = "A set of objects used for comparison."
WINDOWS_DESCRIPTION = "Definition(s) for evaluation of temporal windows"
